from __future__ import annotations

import copy
from dataclasses import dataclass
from typing import Dict, Optional, Union


@dataclass
class CPPParsedTypeInfo:
    '''
    A parsed type, with the type and whether it's a pointer.
    '''
    # The type name (`int`, `vector<float`, etc.)
    name: str

    # Pointer, and how many (2 for `int**`, 0 for `int`, etc.)
    pointer_depth: int

    # True if this is a const variable
    is_const: bool = False

    def __str__(self):
        return self.name + '*' * self.pointer_depth


def parse_type(t_name: str) -> CPPParsedTypeInfo:
    '''Convert a type name string into info for a type

    Args:
        t_name (str): The type name (`float`, `float*`)

    Returns:
        CPPParsedTypeInfo: Parsed info from the type
    '''
    ptr_depth = 0
    while True:
        t_name = t_name.strip()
        if t_name.endswith('*'):
            ptr_depth += 1
            t_name = t_name[:-1]
        else:
            break

    if t_name.startswith('const '):
        is_const = True
        t_name = t_name[6:]
    else:
        is_const = False

    return CPPParsedTypeInfo(t_name, ptr_depth, is_const)


class terminal:
    'Represents something we cannot see inside, like float, or int, or bool'

    def __init__(self, t: Union[str, CPPParsedTypeInfo], p_depth: int = 0, is_const: bool = False):
        '''Create a terminal type - a type that we do not need to see inside

        * int
        * float
        * MyOwnStruct

        Use `p_depth` to represent how indirect to get down to the type (int* is 1, for
        example).)

        Args:
            t (str|CPPParsedTypeInfo): The type to represent
            p_depth (int): How many levels of indirection to get to the type
            is_const (bool): Whether this is a const type
        '''
        if isinstance(t, CPPParsedTypeInfo):
            self._type = t.name
            self._p_depth = t.pointer_depth
            self._is_const = t.is_const
        else:
            self._type = t
            self._p_depth = p_depth
            self._is_const = is_const

    def __str__(self):
        c_str = "const " if self.is_const else ""
        return f"{c_str}{self.type}{'*' * self._p_depth}"

    @property
    def is_a_pointer(self) -> bool:
        'Returns true if this terminal is a pointer'
        return self._p_depth > 0

    @property
    def p_depth(self) -> int:
        'Return how many levels of indirection to get to the type'
        return self._p_depth

    @property
    def is_const(self) -> bool:
        'Returns true if this terminal is a const type'
        return self._is_const

    def default_value(self):
        raise NotImplementedError()

    @property
    def type(self) -> str:
        return self._type

    def get_dereferenced_type(self) -> terminal:
        'Type after dereferencing it once. Will throw if this type cannot be dereferenced'
        if self._p_depth == 0:
            raise RuntimeError(f'Cannot dereference type {self}')

        # Do deep copy because this needs to work in subclasses.
        new_t = copy.copy(self)
        new_t._p_depth -= 1
        return new_t


class collection (terminal):
    'Represents a collection/list/vector of the same type'

    def __init__(self, element_type: terminal, array_type: Optional[Union[str, CPPParsedTypeInfo]] = None, p_depth: int = 0):
        '''Create a collection type, like `vector<float>`.

        Args:
            element_type (terminal): The element type, like a `terminal` of `float`.
            array_type (Optional[Union[str, CPPParsedTypeInfo]], optional): The type of the array. Defaults to None. Everything
                    is lifted from `array_type` if it is a `CPPParsedTypeInfo`.
            p_depth (int, optional): If the array type is a pointer or not. Defaults to 0. Ignored if `array_type` is `CPPParsedTypeInfo`.
        '''
        if array_type is None:
            super().__init__(f"std::vector<{element_type}>", p_depth=p_depth)
        elif isinstance(array_type, CPPParsedTypeInfo):
            super().__init__(array_type)
        else:
            super().__init__(array_type, p_depth=p_depth)

        # And the element type we are representing
        self._element_type = element_type

    # TODO: Turn into a property
    @property
    def element_type(self) -> terminal:
        'The type of element that this collection holds'
        return self._element_type


###########################
# Manage types


@dataclass
class MethodInvokeInfo:
    'Method invocation info'

    # The return type
    r_type: terminal

    # Number of dereferences to apply to the calling object. Normally zero.
    # 0: obj.method()
    # 1: obj->method(), etc.
    deref_depth: int


g_method_type_dict: Dict[str, Dict[str, MethodInvokeInfo]] = {}


def add_method_type_info(type_string: str, method_name: str, t: terminal, deref_depth: int = 0):
    '''
    Define a return type for a method

    type_string         String of the object the method is calling against
    method_name         Name of the object
    t                   The type (terminal, collection, etc.) of return type
    '''
    if type_string not in g_method_type_dict:
        g_method_type_dict[type_string] = {}
    g_method_type_dict[type_string][method_name] = MethodInvokeInfo(t, deref_depth)


def method_type_info(type_string: str, method_name: str) -> Optional[MethodInvokeInfo]:
    '''
    Return the type of the method's return value
    '''
    if type_string not in g_method_type_dict:
        return None
    if method_name not in g_method_type_dict[type_string]:
        return None
    return g_method_type_dict[type_string][method_name]



import ast
from func_adl_xAOD.common.meta_data import generate_script_block
from func_adl_xAOD.atlas.xaod.event_collections import atlas_event_collection_coder, atlas_xaod_collections, define_default_atlas_types
from typing import Any, Callable, Dict
from func_adl_xAOD.common.event_collections import EventCollectionSpecification
from func_adl_xAOD.common.math_utils import get_math_methods
from func_adl_xAOD.atlas.xaod.jets import get_jet_methods
from func_adl_xAOD.atlas.xaod.query_ast_visitor import \
    atlas_xaod_query_ast_visitor
from func_adl_xAOD.common.executor import executor


class atlas_xaod_executor(executor):
    def __init__(self):
        file_names = ['ATestRun_eljob.py', 'package_CMakeLists.txt', 'query.cxx', 'query.h', 'runner.sh']
        runner_name = 'runner.sh'
        template_dir_name = 'func_adl_xAOD/template/atlas/r21'
        self._ecc = atlas_event_collection_coder()
        method_names: Dict[str, Callable[[ast.Call], ast.Call]] = {
            md.name: self.build_callback(self._ecc, md)
            for md in atlas_xaod_collections
        }
        method_names.update(get_jet_methods())
        method_names.update(get_math_methods())
        super().__init__(file_names, runner_name, template_dir_name, method_names)
        define_default_atlas_types()

    @staticmethod
    def build_callback(ecc, md):
        'Required due to by-reference lambda capture not working as expected in python'
        return lambda cd: ecc.get_collection(md, cd)

    def reset(self):
        '''Reset our atlas default types
        '''
        super().reset()
        define_default_atlas_types()

    def get_visitor_obj(self):
        return atlas_xaod_query_ast_visitor()

    def add_to_replacement_dict(self) -> Dict[str, Any]:
        d1 = super().add_to_replacement_dict()

        # Combine metadata script blocks
        d = {
            'job_option_additions': generate_script_block(self._job_option_blocks)
        }
        d.update(d1)
        return d

    def build_collection_callback(self, metadata: EventCollectionSpecification) -> Callable[[ast.Call], ast.Call]:
        '''Build the AST analyzer callback for this collection.

        Args:
            metadata (EventCollectionSpecification): The metadata describing this collection

        Returns:
            Callable[[ast.Call], ast.Call]: Function that will implement what is needed to build the ast properly.
        '''
        if metadata.backend_name != 'atlas':
            raise ValueError(f'Attempt to create a collection from metadata for the {metadata.backend_name} backend; only "atlas" allowed.')

        return lambda cd: self._ecc.get_collection(metadata, cd)



from func_adl_xAOD.common.event_collections import EventCollectionSpecification
from func_adl_xAOD.common.cpp_ast import CPPCodeSpecification
from func_adl_xAOD.common.cpp_types import add_method_type_info, collection, terminal
from func_adl_xAOD.common.cpp_types import CPPParsedTypeInfo, parse_type
from typing import Any, Dict, List, Union
from dataclasses import dataclass, field


@dataclass
class JobScriptSpecification:
    name: str
    script: List[str]
    depends_on: List[str]


@dataclass
class InjectCodeBlock:
    'Code to be directly injected into the hpp and cpp files'

    # Name of the code block
    name: str

    # Include files for the cpp code
    body_includes: List[str] = field(default_factory=list)

    # Include files for the hpp code
    header_includes: List[str] = field(default_factory=list)

    # Instance variable declarations
    private_members: List[str] = field(default_factory=list)

    # Instance variable ctor initializers
    instance_initialization: List[str] = field(default_factory=list)

    # Code lines to place in the constructor
    ctor_lines: List[str] = field(default_factory=list)

    # Lines to add to initialize statement
    initialize_lines: List[str] = field(default_factory=list)

    # Packages/Libraries to add to the CMake lib line
    link_libraries: List[str] = field(default_factory=list)


SpecificationTypes = Union[CPPCodeSpecification, EventCollectionSpecification, JobScriptSpecification, InjectCodeBlock]


def ok_to_add_code_block(spec, cpp_funcs: List[SpecificationTypes]) -> bool:
    '''Check already added code blocks specs to see if there is one with the same name,
    and if so, if its contents are the same.

    Return ok if this is a unique code block
    Return false if this is a duplicate
    Throw an error if the name but not content matches.

    Args:
        spec (InjectCodeBlock): The block to be going after
        cpp_funcs ([type]): The list code blocks we should find a match in
    '''
    for b in cpp_funcs:
        if isinstance(b, InjectCodeBlock) and b.name == spec.name:
            if b == spec:
                return False
            raise ValueError(f'Duplicate inject_code blocks with name {spec.name} that are not identical. Do not know which one to use! first: {b} second: {spec}')
    return True


def process_metadata(md_list: List[Dict[str, Any]]) -> List[SpecificationTypes]:
    '''Process a list of metadata, in order.

    Args:
        md (List[Dict[str, str]]): The metadata to process

    Returns:
        List[X]: Metadata we've found
    '''
    cpp_funcs: List[SpecificationTypes] = []
    for md in md_list:
        md_type = md.get('metadata_type')
        if md_type is None:
            raise ValueError(f'Metadata is missing `metadata_type` info ({md})')

        if md_type == 'add_method_type_info':
            if 'return_type' in md:
                # Single return type
                type_info = parse_type(md['return_type'])
                term = terminal(type_info.name, p_depth=type_info.pointer_depth)
            else:
                type_info_element = parse_type(md['return_type_element'])
                type_info_collection = parse_type(md['return_type_collection']) if 'return_type_collection' in md else CPPParsedTypeInfo(f'std::vector<{type_info_element}>', 0)
                term = collection(terminal(type_info_element), array_type=type_info_collection)
            d_count = 0
            if 'deref_count' in md:
                d_count = int(md['deref_count'])
            add_method_type_info(md['type_string'], md['method_name'], term, d_count)
        elif md_type == 'inject_code':
            info = dict(md)
            del info['metadata_type']
            if len(info) > 0:
                try:
                    spec = InjectCodeBlock(**info)
                except TypeError as e:
                    raise ValueError(f'Bad inject_code block item: {str(e)}')
                if ok_to_add_code_block(spec, cpp_funcs):
                    cpp_funcs.append(spec)
        elif md_type == 'add_job_script':
            spec = JobScriptSpecification(
                name=md['name'],
                script=md['script'],
                depends_on=md.get('depends_on', [])
            )
            cpp_funcs.append(spec)
        elif md_type == 'add_cpp_function':
            spec = CPPCodeSpecification(
                md['name'],
                md['include_files'],
                md['arguments'],
                md['code'],
                md['result_name'] if 'result_name' in md else 'result',
                md['return_type'],
                bool(md['return_is_collection']) if 'return_is_collection' in md else False,
                md['method_object'] if 'method_object' in md else None,
                md['instance_object'] if 'instance_object' in md else None,
            )
            cpp_funcs.append(spec)
        elif md_type == 'add_atlas_event_collection_info':
            for k in md.keys():
                if k not in ['metadata_type', 'name', 'include_files', 'container_type', 'element_type', 'contains_collection', 'link_libraries']:
                    raise ValueError(f'Unexpected key {k} when declaring ATLAS collection metadata')
            if (md['contains_collection'] and 'element_type' not in md) or (not md['contains_collection'] and 'element_type' in md):
                raise ValueError('In collection metadata, `element_type` must be specified if `contains_collection` is true and not if it is false')

            from func_adl_xAOD.atlas.xaod.event_collections import atlas_xaod_event_collection_collection, atlas_xaod_event_collection_container
            container_type = atlas_xaod_event_collection_collection(md['container_type'], md['element_type']) if md['contains_collection'] \
                else atlas_xaod_event_collection_container(md['container_type'])
            link_libraries = [] if 'link_libraries' not in md else md['link_libraries']
            spec = EventCollectionSpecification(
                'atlas',
                md['name'],
                md['include_files'],
                container_type,
                link_libraries)
            cpp_funcs.append(spec)
        elif md_type == 'add_cms_event_collection_info':
            for k in md.keys():
                if k not in ['metadata_type', 'name', 'include_files', 'container_type', 'element_type', 'contains_collection', 'element_pointer']:
                    raise ValueError(f'Unexpected key {k} when declaring ATLAS collection metadata')
            if (md['contains_collection'] and 'element_type' not in md) or (not md['contains_collection'] and 'element_type' in md):
                raise ValueError('In collection metadata, `element_type` must be specified if `contains_collection` is true and not if it is false')

            from func_adl_xAOD.cms.aod.event_collections import cms_aod_event_collection_collection
            container_type = cms_aod_event_collection_collection(md['container_type'], md['element_type'])
            # container_type = cms_aod_event_collection_collection(md['container_type'], md['element_type']) if md['contains_collection'] \
            #     else cms_aod_event_collection_container(md['container_type'])

            spec = EventCollectionSpecification(
                'cms',
                md['name'],
                md['include_files'],
                container_type,
                [])
            cpp_funcs.append(spec)
        else:
            raise ValueError(f'Unknown metadata type ({md_type})')

    return cpp_funcs


def generate_script_block(blocks: List[JobScriptSpecification]) -> List[str]:
    '''Returns the script block to insert into the job control.

    * Takes dependencies into account
    * Gets rid of any duplicates
    * Will combine any blocks with the same script text, but different dependencies.

    Args:
        blocks (List[JobScriptSpecification]): The list of, unordered, dependency blocks

    Returns:
        List[str]: The list of insertions to insert.
    '''
    # Build the dependency graph and check for improper
    # duplications and dependencies.
    dependencies: Dict[str, List[str]] = {}
    block_lookup: Dict[str, JobScriptSpecification] = {}

    for b in blocks:
        if b.name not in dependencies:
            dependencies[b.name] = []
            block_lookup[b.name] = b
        else:
            if b.script != block_lookup[b.name].script:
                raise ValueError(f'Duplicate metadata block "{b.name}", but blocks are not identical ({b.script} and {block_lookup[b.name].script} should be identical)!')

        dependencies[b.name].extend(b.depends_on)

    for name, deps in dependencies.items():
        for d in deps:
            if d not in dependencies:
                raise ValueError(f'Dependent metadata block {d} not found in sent metadata (from {name}!')

    # Next, start from blocks that have no dependencies and work our way up.
    seen_blocks = set()
    script_text = []

    while len(seen_blocks) < len(dependencies):
        emitted = False
        for j in block_lookup.values():
            if j.name not in seen_blocks:
                if set(dependencies[j.name]) <= seen_blocks:
                    for ln in j.script:
                        script_text.append(ln)
                    seen_blocks.add(j.name)
                    emitted = True
        if not emitted:
            remaining_blocks = ', '.join((set(block_lookup.keys()) - seen_blocks))
            raise ValueError(f'There seems to be a metadata script block circular dependency ({remaining_blocks})')

    return script_text
